//
//  BLEAPI.swift
//  qcc5100.sample
//
//

import Foundation
import CoreBluetooth

enum Service: CaseIterable {
    case deviceInfo
    case peripheralRiiiver
}

extension Service: RawRepresentable {
    init?(rawValue: CBUUID) {
        guard let c = { () -> Service? in
            for iCase in Service.allCases {
                if rawValue == iCase.rawValue {
                    return iCase
                }
            }
            return nil
        }() else { return nil }
        self = c
    }
    
    var rawValue: CBUUID {
        switch self {
        case .deviceInfo: return CBUUID(string: "180A")
        case .peripheralRiiiver: return CBUUID(string: "0000A000-0000-1000-8000-00805F9B34FB")
        }
    }
}

enum Characteristic: CaseIterable {
    case onBoardUserEvent
    case commandReceiver
}

extension Characteristic: RawRepresentable {
    init?(rawValue: CBUUID) {
        guard let c = { () -> Characteristic? in
            for iCase in Characteristic.allCases {
                if rawValue == iCase.rawValue {
                    return iCase
                }
            }
            return nil
        }() else { return nil }
        self = c
    }
    
    var rawValue: CBUUID {
        switch self {
        case .onBoardUserEvent: return CBUUID(string: "0000A001-0000-1000-8000-00805F9B34FB")
        case .commandReceiver: return CBUUID(string: "0000A002-0000-1000-8000-00805F9B34FB")
        }
    }
}

protocol Command {
    var name: String { get }
}

enum OnBoardUserEventCommand: UInt8, CaseIterable {
    case button1press = 0
    case button2press = 1
    case success      = 2
    case failure      = 3
}

extension OnBoardUserEventCommand: Command {
    var name: String {
        switch self {
        case OnBoardUserEventCommand.button1press:
            return "button1press"
        case OnBoardUserEventCommand.button2press:
            return "button2press"
        case OnBoardUserEventCommand.success:
            return "success"
        case OnBoardUserEventCommand.failure:
            return "failure"
        }
    }
    
    var responseCommand: CommandReceiverCommand? {
        switch self {
        case OnBoardUserEventCommand.button1press:
            return CommandReceiverCommand.button1result
        case OnBoardUserEventCommand.button2press:
            return CommandReceiverCommand.button2result
        default:
            return nil
        }
    }
}

enum CommandReceiverCommand: UInt8, CaseIterable {
    case button1result = 0
    case button2result = 1
    case display       = 2
    case success       = 3
    case failure       = 4
}

extension CommandReceiverCommand: Command {
    var name: String {
        switch self {
        case CommandReceiverCommand.button1result:
            return "button1result"
        case CommandReceiverCommand.button2result:
            return "button2result"
        case CommandReceiverCommand.display:
            return "display"
        case CommandReceiverCommand.success:
            return "success"
        case CommandReceiverCommand.failure:
            return "failure"
        }
    }
}

enum Operation {
    case toggleScanState
    case startScan
    case stopScan
    case connect(uuid: String)
    case cancelConnect
    case discoverService
    case writeDataP2W(message: String)
}

protocol BLEOperation {
    var rawValue: String { get }
    var operation: Operation { get }
    var delegate: BLECentralDelegate? { get }
    var command: CommandReceiverCommand? { get }
}

final class BLEAPI {
    
    struct ToggleScanState: BLEOperation {
        var rawValue: String = "ToggleScanState"
        var operation: Operation = .toggleScanState
        var delegate: BLECentralDelegate? = nil
        var command: CommandReceiverCommand? = nil
    }
    
    struct StartScan: BLEOperation {
        var rawValue: String = "StartScan"
        var operation: Operation = .startScan
        var delegate: BLECentralDelegate? = nil
        var command: CommandReceiverCommand? = nil
    }
    
    struct StopScan: BLEOperation {
        var rawValue: String = "StopScan"
        var operation: Operation = .stopScan
        var delegate: BLECentralDelegate? = nil
        var command: CommandReceiverCommand? = nil
    }
    
    struct Connect: BLEOperation {
        let uuid: String
        var rawValue: String = "Connect"
        var operation: Operation {
            return .connect(uuid: self.uuid)
        }
        var delegate: BLECentralDelegate?
        var command: CommandReceiverCommand? = nil

        init(uuid: String, delegate: BLECentralDelegate? = nil) {
            self.uuid = uuid
            self.delegate = delegate
        }
    }
    
    struct CancelConnect: BLEOperation {
        var rawValue: String = "CancelConnect"
        var operation: Operation = .cancelConnect
        var command: CommandReceiverCommand? = nil
        var delegate: BLECentralDelegate? = nil
    }
    
    struct DiscoverService: BLEOperation {
        var rawValue: String = "DiscoverService"
        var operation: Operation = .discoverService
        var delegate: BLECentralDelegate?
        var command: CommandReceiverCommand? = nil
        
        init(delegate: BLECentralDelegate? = nil) {
            self.delegate = delegate
        }
    }


    struct SendResultButton1: BLEOperation {
        let message: String

        var rawValue: String = "SendResultButton1"
        
        var operation: Operation {
            return .writeDataP2W(message: self.message)
        }
        
        var command: CommandReceiverCommand? = .button1result
        var delegate: BLECentralDelegate? = nil
        
        init(message: String) {
            self.message = message
        }
    }
    
    struct SendResultButton2: BLEOperation {
        let message: String
        
        var rawValue: String = "SendResultButton2"
        
        var operation: Operation {
            return .writeDataP2W(message: self.message)
        }
        
        var command: CommandReceiverCommand? = .button2result
        var delegate: BLECentralDelegate? = nil
        
        init(message: String) {
            self.message = message
        }
    }
    
    struct SendDisplayData: BLEOperation {
        let message: String
        
        var rawValue: String = "SendDisplayData"
        
        var operation: Operation {
            return .writeDataP2W(message: self.message)
        }
        
        var command: CommandReceiverCommand? = .display
        var delegate: BLECentralDelegate? = nil
        
        init(message: String) {
            self.message = message
        }
    }
}
