//
//  BLECentralClient.swift
//  qcc5100.sample
//

import Foundation
import CoreBluetooth

protocol BLECentralRequestable: AnyObject {
    func operate<Request: BLEOperation>(request: Request)
    func clearAllFounded()
}

class BLECentralClient: NSObject {
    private var isScanning: Bool = false
    private var manager: CBCentralManager!
    private var delegates: [BLECentralDelegate] = []
    private(set) var entity: BLEEntity!
    
    override private init() {
        super.init()
        let options: [String : Any] = [
            CBCentralManagerOptionShowPowerAlertKey: true,
            CBCentralManagerOptionRestoreIdentifierKey: "RIIIVER_QCC5100_CENTRAL"
        ]
        manager = CBCentralManager(delegate: self, queue: nil, options: options)
        entity = BLEEntity()
    }
    
    static let sharedInstance = BLECentralClient()
    
    private func toggleScanState() {
        if isScanning {
            stopScan()
        } else {
            startScan()
        }
    }
    
    private func startScan() {
        if self.isScanning {
            print("[BLECentralClient] startScan: already scanning")
        } else {
            print("[BLECentralClient] startScan: start")
            self.isScanning = true
            self.manager.scanForPeripherals(withServices: nil)
        }
    }
    
    private func stopScan() {
        if !isScanning {
            print("[BLECentralClient] stopScan: already stop")
            return
        }
        print("[BLECentralClient]: stopScan: stop")
        self.isScanning = false
        self.manager.stopScan()
    }
    
    private func add(_ peripheral: PeripheralEntity) {
        var currentPeripherals = entity.foundPeripherals
        if let index = currentPeripherals.firstIndex(where: { $0.uuid.uuidString == peripheral.uuid.uuidString }) {
            currentPeripherals[index] = peripheral
        } else {
            currentPeripherals.append(peripheral)
        }
        entity.foundPeripherals = currentPeripherals
    }
    
    private func connect(to uuid: String, delegate: BLECentralDelegate? = nil) {
        guard let target = self.entity.foundPeripherals.first(where: { $0.uuid.uuidString == uuid }),
              let peripheral = target.rawInstance else { return }

        if let delegate,
           !self.delegates.contains(where: { $0.identifier == delegate.identifier }) {
            self.delegates.append(delegate)
        }
        self.manager.connect(peripheral)
    }
    
    private func cancelConnect() {
        guard let connected = self.entity.currentConnected,
              let peripheral = connected.rawInstance else { return }
        self.manager.cancelPeripheralConnection(peripheral)
    }
    
    private func discoverService(delegate: BLECentralDelegate? = nil) {
        guard let connected = self.entity.currentConnected,
              let peripheral = connected.rawInstance else { return }

        if let delegate,
           !self.delegates.contains(where: { $0.identifier == delegate.identifier }) {
            self.delegates.append(delegate)
        }
        peripheral.delegate = self
        peripheral.discoverServices(nil)
    }
    
    private func handleSlateRequest(data: Data) {
        let payload = data.slateRiiiverFormat
        guard let command = OnBoardUserEventCommand.init(rawValue: payload.commandId) else { return }
        print("## received command: \(command.name)")
        
        switch command {
        case .button1press: fallthrough
        case .button2press:
            // 新たなW2Pボタン押しコマンドを受け付けた際に実行待ちのP2Wコマンドがある場合はアーリーリターンする
            // FIXME: thrrow some exception due to multiple execution/request
            guard self.entity.phoneToWatchQue == nil else { return }
            // 受け付けたコマンドに対応する応答コマンドをEntityに保持する
            guard let respCommand = command.responseCommand else { return }
            self.entity.phoneToWatchQue = respCommand
            // 受け付けたコマンドに対応する実行対象のiiideaIdが指定されていない場合アーリーリターンする
            guard let iiideaId = RiiiverSDKClient.sharedInstance.entity.slateAppletIds[respCommand] as? String else {
                print("## iiidea is not set for this command")
                return
            }
            // Kick the trigger
            QCC5100TriggerPieceCore.didTrigger(appletId: iiideaId, input: nil)
        case .success: fallthrough
        case .failure: fallthrough
        default:
            // receive response of success or failure == end of sequence
            self.entity.phoneToWatchQue = nil
            // Send app's internal notification for receving the resposne of iiidea execution
            var userInfo: [String:Any] = [:]
            userInfo[BlockNotification.Key.Result] = (command == .success)
            NotificationCenter.default.post(name: BlockNotification.Name.ActionResponse, object: nil, userInfo: userInfo)
        }
    }
    
    private func genPayload(message: String, commandId: CommandReceiverCommand) -> Data {
        // the max value represented with 2 bytes unsigned interger is 65535
        // TODO: throw some error if overflow
        // dataLength: 2 bytes, commandId: 1 byte, message: n bytes, where dataLength is not including itself.
        let length = (message.lengthOfBytes(using: .utf8) + 1).data.subdata(of: 0, length: 2)

        let payload = SlateRiiiverPayload(dataLength: length, commandId: commandId.rawValue, message: [UInt8](message.utf8))
        print("## payload: \(payload.combinedByteArray)")
        return payload.combinedData
    }
    
    private func write(data: Data) {
        print("[BLECentralClient] write: in")

        guard let periheral = entity.currentConnected?.rawInstance,
              let characteristic = entity.currentConnected?.commandReceiverCharacteristic else { return }
                
        periheral.writeValue(data, for: characteristic, type: .withResponse)
    }
}

extension BLECentralClient: BLECentralRequestable {
    func operate<Request>(request: Request) where Request : BLEOperation {
        switch request.operation {
        case .toggleScanState:
            toggleScanState()
        case .startScan:
            startScan()
        case .stopScan:
            stopScan()
        case .connect(let uuid):
            connect(to: uuid, delegate: request.delegate)
        case .cancelConnect:
            cancelConnect()
        case .discoverService:
            discoverService(delegate: request.delegate)
        case .writeDataP2W(let message):
            write(data: genPayload(message: message, commandId: request.command!))
        }
    }
    
    func clearAllFounded() {
        self.entity.foundPeripherals = []
    }
}

extension BLECentralClient: CBCentralManagerDelegate {
    func centralManagerDidUpdateState(_ central: CBCentralManager) {
        // TODO: implement later
    }
    
    func centralManager(_ central: CBCentralManager, willRestoreState dict: [String : Any]) {
        // TODO: implement later
    }
    
    func centralManager(_ central: CBCentralManager, didDiscover peripheral: CBPeripheral, advertisementData: [String : Any], rssi RSSI: NSNumber) {
        print("[BLECentralDelegate] didDiscover: UUID:\(peripheral.identifier), peripheral:\(peripheral), advertisementData:\(advertisementData)")

        let foundPeripheral = PeripheralEntity(name: peripheral.name ?? advertisementData["kCBAdvDataLocalName"] as? String ?? "NO_NAME",
                                               uuid: peripheral.identifier,
                                               advertisementData: advertisementData,
                                               rssi: RSSI.intValue,
                                               rawInstance: peripheral)
        
        if foundPeripheral.name == "NO_NAME" {
            return
        }
        
        add(foundPeripheral)
    }
    
    func centralManager(_ central: CBCentralManager, didConnect peripheral: CBPeripheral) {
        print("[BLECentralDelegate] didConnect: in")
        
        if let known = self.entity.foundPeripherals.first(where: { $0.uuid == peripheral.identifier }) {
            self.entity.currentConnected = known
            self.delegates.forEach { $0.didConnectPeripheral?(peripheral) }
        } else {
            let connected = PeripheralEntity(name: peripheral.name ?? "NO_NAME",
                                             uuid: peripheral.identifier,
                                             advertisementData: nil,
                                             rssi: nil,
                                             rawInstance: peripheral)
            self.entity.currentConnected = connected
            self.delegates.forEach { $0.didConnectPeripheral?(peripheral) }
        }
    }
    
    func centralManager(_ central: CBCentralManager, didFailToConnect peripheral: CBPeripheral, error: Error?) {
        // TODO: implement later
    }
    
    func centralManager(_ central: CBCentralManager, didDisconnectPeripheral peripheral: CBPeripheral, error: Error?) {
        print("[BLECentralDelegate] didDisconnect: in")
        self.entity.currentConnected = nil
    }
}

extension BLECentralClient: CBPeripheralDelegate {
    // Invoked when services are found
    func peripheral(_ peripheral: CBPeripheral, didDiscoverServices error: Error?) {
        print("[CBPeripheralDelegate] didDiscoverServices: in")
        guard let services = peripheral.services,
              services.contains(where: { $0.uuid == Service.peripheralRiiiver.rawValue }) else {
            self.delegates.forEach { $0.didNotFoundService?() }
            self.delegates = []
            return
        }
        self.delegates.forEach { $0.didFoundService?() }
        self.delegates = []
        services
            .filter { $0.uuid == Service.peripheralRiiiver.rawValue }
            .forEach { service in
                peripheral.discoverCharacteristics(Characteristic.allCases.map { $0.rawValue }, for: service)
            }
    }
    
    // Invoked when characteristics are found
    func peripheral(_ peripheral: CBPeripheral, didDiscoverCharacteristicsFor service: CBService, error: Error?) {
        print("[CBPeripheralDelegate] didDiscoverCharacteristic: in")
        for characteristic in service.characteristics! {
            if characteristic.uuid == Characteristic.onBoardUserEvent.rawValue {
                print("## Characteristic:W2PC found")
                self.entity.currentConnected?.onBoardUserEventCharacteristic = characteristic
                peripheral.setNotifyValue(true, for: characteristic)
            } else if characteristic.uuid == Characteristic.commandReceiver.rawValue {
                print("## Characteristic:P2WC found")
                self.entity.currentConnected?.commandReceiverCharacteristic = characteristic
            }
        }
    }
    
    // Invoked when receiving the response of write
    func peripheral(_ peripheral: CBPeripheral, didWriteValueFor characteristic: CBCharacteristic, error: Error?) {
        if let error {
            print("[CBPeripheralDelegate] didWriteValueFor: error!\n## \(error.localizedDescription)")
            return
        }
        print("[CBPeripheralDelegate] didWriteValueFor: no error")
    }
    
    // Invoked when notification state updates
    func peripheral(_ peripheral: CBPeripheral, didUpdateNotificationStateFor characteristic: CBCharacteristic, error: Error?) {
        print("[CBPeripheralDelegate] didUpdateNotificationStateFor: invoked")
        if let error {
            print("[CBPeripheralDelegate] didUpdateNotificationStateFor: error!\n## \(error.localizedDescription)")
            return
        }
        switch characteristic.uuid {
        case Characteristic.onBoardUserEvent.rawValue:
            print("## W2PC notification state updated: \(characteristic.isNotifying)")
        case Characteristic.commandReceiver.rawValue:
            print("## P2WC notification state updated: \(characteristic.isNotifying)")
        default:
            print("## unknown characteristic updated")
        }
    }
    
    // Invoked when updating data of Read/Notify
    func peripheral(_ peripheral: CBPeripheral, didUpdateValueFor characteristic: CBCharacteristic, error: Error?) {
        print("[CBPeripheralDelegate] didUpdateValueFor: invoked")
        if let error {
            print("[CBPeripheralDelegate] didUpdateValueFor: error!\n## \(error.localizedDescription)")
            return
        }
        guard let data = characteristic.value else { return }

        switch characteristic.uuid {
        case Characteristic.onBoardUserEvent.rawValue:
            print("## W2PC updated: \(data.hexString)")
            handleSlateRequest(data: data)
        case Characteristic.commandReceiver.rawValue:
            print("## P2WC updated: \(data.hexString)")
        default:
            print("## unknown characteristic updated")
        }
    }
}
