//
//  BLECentralDelegate.swift
//  qcc5100.sample
//

import CoreBluetooth

/// BLEセントラルのプロトコル定義。
@objc protocol BLECentralDelegate {
    @objc var identifier: String { get }

    /// Bluetoothの状態更新時に呼ばれる。
    ///
    /// - Parameters:
    ///   - state: Bluetooth状態を指定する
    @objc optional func didUpdateState(_ state: CBManagerState)

    /// スキャン開始時に呼ばれる。
    ///
    /// - Parameters:
    ///   - services: スキャン開始時のサービスを指定する
    @objc optional func didStartScan(services: [CBUUID])
    
    /// スキャン停止時に呼ばれる。
    ///
    /// - Parameters:
    ///   - reconnect: スキャン停止後に接続済みデバイスに対して接続要求を行う場合はtrue
    @objc optional func didStopScan(reconnect: Bool)
    
    /// スキャンによりペリフェラルを検出した際に呼ばれる。
    ///
    /// - Parameters:
    ///   - peripheral: 検出したペリフェラルを指定する
    ///   - advertisementData: 検出したペリフェラルのアドバタイズデータを指定する
    @objc optional func didDiscoverPeripheral(_ peripheral: CBPeripheral, advertisementData: [String:Any])
    
    /// ペリフェラルと接続した際に呼ばれる。
    /// NOTE: OS的に接続した通知。CDIDなどの接続情報はこの時点では確定していないので注意。
    ///
    /// - Parameters:
    ///   - peripheral: 接続したペリフェラルを指定する
    @objc optional func didConnectPeripheral(_ peripheral: CBPeripheral)

    /// 接続要求をキャンセルした
    @objc optional func didCancelConnectPeripheral(_ peripheral: CBPeripheral)
    
    /// ペリフェラルとの接続に失敗した際に呼ばれる。
    ///
    /// - Parameters:
    ///   - peripheral: 接続に失敗したペリフェラルを指定する
    ///   - error: エラー情報を指定する
    @objc optional func didFailToConnectPeripheral(_ peripheral: CBPeripheral, error: Error?)

    /// ペリフェラルと切断した際に呼ばれる。
    /// または、接続要求のキャンセルをした際に呼ばれる。
    ///
    /// - Parameters:
    ///   - peripheral: 切断したペリフェラルを指定する
    ///   - error: エラー情報を指定する
    @objc optional func didDisconnectPeripheral(_ peripheral: CBPeripheral, error: Error?)

    /// 接続したペリフェラルにSlateRiiiverサービスが存在しない場合に呼ばれる
    @objc optional func didNotFoundService()

    /// 接続したペリフェラルにSlateRiiiverサービスが存在した場合に呼ばれる
    @objc optional func didFoundService()
    
    /// 状態の保存と復元時に呼ばれる。
    ///
    /// - Parameters:
    ///   - state: 状態を指定する
    @objc optional func willRestoreState(_ state: [String : Any])

    /// ペリフェラルからのインジケーション設定結果を受けた際に呼ばれる。
    ///
    /// - Parameters:
    ///   - peripheral: インジケーション設定結果を受けたペリフェラルを指定する
    ///   - characteristic: インジケーション設定結果を受けたキャラクタリスティックを指定する
    ///   - error: エラー情報を指定する
    @objc optional func didUpdateNotificationStateForPeripheral(_ peripheral: CBPeripheral, characteristic: CBCharacteristic, error: Error?)
}
