//
//  PermissionManager.swift
//  qcc5100.sample
//

import Foundation
import CoreLocation
import MediaPlayer
import RiiiverSDK

protocol IPermissionRequest {
    var rawValue: String { get }
    var permissionType: RiiiverSDK.ERPermissionType { get }
}

final class PermissionRequest {
    struct Notification: IPermissionRequest {
        var rawValue: String = "notification"
        var permissionType: RiiiverSDK.ERPermissionType = .notification
    }

    struct Location: IPermissionRequest {
        var rawValue: String = "location"
        var permissionType: RiiiverSDK.ERPermissionType = .location
    }
    
    struct MediaLibrary: IPermissionRequest {
        var rawValue: String = "mediaLibrary"
        var permissionType: RiiiverSDK.ERPermissionType = .mediaLibrary
    }
    
    struct Calendar: IPermissionRequest {
        var rawValue: String = "calendar"
        var permissionType: RiiiverSDK.ERPermissionType = .calendar
    }
    
    struct Microphone: IPermissionRequest {
        var rawValue: String = "microphone"
        var permissionType: RiiiverSDK.ERPermissionType = .microphone
    }
}

protocol PermissionManagerInterface: AnyObject {
    func request<Request: IPermissionRequest>(request: Request)
}

class PermissionManager: NSObject {
    /// Singleton
    public static let sharedInstance = PermissionManager()
    private(set) var entity: PermissionEntity!
    private var locationManager: CLLocationManager!
    
    private init(entity: PermissionEntity = PermissionEntity(), locationManager: CLLocationManager = CLLocationManager()) {
        super.init()
        print("ERPermissionManager init()")
        self.entity = entity
        self.locationManager = locationManager
        self.locationManager.delegate = self
    }
    
    // 位置情報アクセス権限要求
    private func requestForLocation() {
        let stauts = locationManager.authorizationStatus
        switch stauts {
        case .notDetermined:
            locationManager.requestAlwaysAuthorization()
        case .restricted: fallthrough
        case .denied:
            guard let settingsURL = URL(string: UIApplication.openSettingsURLString ) else {
                return
            }
            UIApplication.shared.open(settingsURL, options: [:], completionHandler: nil)
        case .authorizedAlways: fallthrough
        case .authorizedWhenInUse:
            break
        @unknown default:
            break
        }
    }
    
    // メディアライブラリーアクセス権限要求
    private func requestForMediaLibrary() {
        let status = MPMediaLibrary.authorizationStatus()
        switch status {
        case .notDetermined:
            MPMediaLibrary.requestAuthorization() { status in
                DispatchQueue.main.async {
                    self.entity.mediaLibrary = status
                }
            }
        case .restricted: fallthrough
        case .denied:
            guard let settingsURL = URL(string: UIApplication.openSettingsURLString ) else {
                return
            }
            UIApplication.shared.open(settingsURL, options: [:], completionHandler: nil)
            break
        case .authorized:
            break
        @unknown default:
            break
        }
    }
    
    private func requestForNotification() {
        let center = UNUserNotificationCenter.current()
        center.getNotificationSettings { settings in
            let status = settings.authorizationStatus
            switch status {
            case .notDetermined:
                UNUserNotificationCenter.current().requestAuthorization(options: [.alert, .sound]) { (result, error) in
                    if let error {
                        print("## requestForNotification failed due to \(error.localizedDescription)")
                    }
                    if result {
                        DispatchQueue.main.async {
                            self.entity.notification = .authorized
                        }
                    }
                }
                break
            case .provisional: fallthrough
            case .ephemeral: fallthrough
            case .denied:
                guard let settingsURL = URL(string: UIApplication.openSettingsURLString ) else {
                    return
                }
                UIApplication.shared.open(settingsURL, options: [:], completionHandler: nil)
                break
            case .authorized: fallthrough
            @unknown default:
                break
            }
        }
    }
    
    private func requestForCalendar() {
        // TODO: implement later
    }
    
    private func requestForMicrophone() {
        // TODO: implement later
    }

}

extension PermissionManager: PermissionManagerInterface {
    func request<Request>(request: Request) where Request : IPermissionRequest {
        switch request.permissionType {
        case .auth:
            break
        case .calendar:
            requestForCalendar()
        case .location:
            requestForLocation()
        case .mediaLibrary:
            requestForMediaLibrary()
        case.microphone:
            requestForMicrophone()
        case .notification:
            requestForNotification()
        @unknown default:
            break
        }
    }
}

extension PermissionManager: CLLocationManagerDelegate {    
    func locationManagerDidChangeAuthorization(_ manager: CLLocationManager) {
        self.entity.location = manager.authorizationStatus
    }
}
