//
//  RiiiverSDKClient.swift
//  qcc5100.sample
//

import Foundation
import RiiiverSDK

protocol ERAuthClientInterface: AnyObject {
    // ERAuth
    var isLoggedIn: Bool { get }
    func getLoginUrl() async throws -> URL
    func verifyCode(of url: URL) async throws -> Void
    func logout() async throws -> Void
    func setNotificationToken(_ fcmToken: String) async throws -> Void
}

protocol RiiiverSDKClientInterface: AnyObject {
    // ERSDK
    func getMyApplets() async throws -> Void
    func removeMyApplet(_ iiideaId: String) async throws -> Void
    func getERAppletInstance(of iiideaId: String) -> ERApplet?
    func changeTargetIiidea(of: CommandReceiverCommand, iiideaId: String?)
    func clearMyApplets()
    func executeTaskForRemoteNotification(_ remoteNotification: [AnyHashable: Any]) async throws -> Void
}

class RiiiverSDKClient: NSObject {
    private init(entity: RiiiverAppletsEntity = RiiiverAppletsEntity.sharedInstance) {
        super.init()
        self.entity = entity
    }
    
    static let sharedInstance = RiiiverSDKClient()

    private(set) var entity: RiiiverAppletsEntity!
    
    private func initialize() async throws -> Void {
        let error = await RiiiverSdkManager.initialize(deviceIdList: RiiiverConstants.deviceIds)
        if let error {
            print("## RiiiverSdkClient: initialized failed due to \(error.localizedDescription)")
            throw error
        }
        RiiiverSdkManager.addListener(listener: self)
        return
    }
    
    private func forcedDownloadApplets() async throws -> Void {
        guard RiiiverConstants.forcedDownloadAppletIds.count > 0 else {
            print("forcedDownloadApplets: no need to start because there are no targets")
            return
        }
        
        let result = await RiiiverSdkManager.forcedDownloadApplet(appletIds: RiiiverConstants.forcedDownloadAppletIds)
        if let error = result.error {
            print("forcedDownloadApplets failed due to: \(error)")
            throw error
        } else {
            return
        }
    }
}

extension RiiiverSDKClient: ERAuthClientInterface {
    var isLoggedIn: Bool {
        return RiiiverSdkManager.authenticated
    }
    
    func getLoginUrl() async throws -> URL {
        try await self.initialize()
        
        // Get the URL for login.
        let result = await RiiiverSdkManager.getLoginUrl()
        if let error = result.error {
            throw error
        }
        guard let loginUrl = result.url else {
            throw ERError(ERErrorCode.noServerUrl)
        }
        return loginUrl
    }
    
    func verifyCode(of url: URL) async throws -> Void {
        try await self.initialize()
        
        // Check whether the query param "token", which behave like authorization_code, presents in the redirect URL.
        // If presents, store as property of ERAuth class.
        if let error = await RiiiverSdkManager.verifyCodeOf(uri: url) {
            throw error
        }
        return
    }
    
    func logout() async throws -> Void {
        try await self.initialize()
        
        let error = await RiiiverSdkManager.logout()
        if let error {
            print("## logout failed due to \(error)")
            throw error
        }
        return
    }
    
    func setNotificationToken(_ fcmToken: String) async throws {
        try await self.initialize()

        let error = await RiiiverSdkManager.setNotificationToken(notificationToken: fcmToken)
        if let error {
            print("## setNotification failed due to \(error)")
            throw error
        }
        return
    }
}

extension RiiiverSDKClient: RiiiverSDKClientInterface {
    func getMyApplets() async throws -> Void {
        try await self.initialize()

        try await self.forcedDownloadApplets()
        
        let result = await RiiiverSdkManager.getMyApplets(deviceIdList: RiiiverConstants.deviceIds)
        if let error = result.error {
            print("getMyApplets end with some error: \(error)")
            throw error
        } else {
            DispatchQueue.main.async {
                self.entity.myApplets = result.iiideas
            }
        }
        return
    }
    
    func removeMyApplet(_ iiideaId: String) async throws -> Void {
        try await self.initialize()

        guard let iiidea = self.entity.myApplets?.first(where: { $0.getId() == iiideaId}) else {
            print("## the target iiidea not found")
            return
        }
        
        let result = await RiiiverSdkManager.removeMyIiidea(iiidea: iiidea)
        if let error = result.error {
            print("## removeMyApplet failed due to: \(error)")
            throw error
        } else {
            print("## \(iiideaId) has been removed successfully")
            return
        }
    }
    
    func getERAppletInstance(of iiideaId: String) -> ERApplet? {
        guard let iiidea = self.entity.myApplets?.first(where: { $0.getId() == iiideaId }) else {
            return nil
        }
        return iiidea
    }
    
    func changeTargetIiidea(of: CommandReceiverCommand, iiideaId: String?) {
        DispatchQueue.main.async {
            self.entity.slateAppletIds[of] = iiideaId
        }
    }
    
    func clearMyApplets() {
        DispatchQueue.main.async {
            self.entity.clearMyApplets()
        }
    }
    
    func executeTaskForRemoteNotification(_ remoteNotification: [AnyHashable: Any]) async throws -> Void {
        try await self.initialize()

        let error = await RiiiverSdkManager.executeTaskForRemoteNotification(userInfo: remoteNotification)
        if let error {
            print("## executeTaskForRemoteNotification is failed due to \(error)")
            throw error
        }
        return
    }
}

extension RiiiverSDKClient: ERSdkListener {
    func onLogin(applets: Array<RiiiverSDK.ERApplet>) {
        print("## ERSDKDelegate onLogin: in")
        DispatchQueue.main.async {
            if let delegate = UIApplication.shared.delegate as? AppDelegate,
               let fcmToken = delegate.fcmToken {
                Task(priority: .medium) { [weak self] in
                    guard let self else { return }
                    do {
                        try await self.setNotificationToken(fcmToken)
                    } catch (let error) {
                        print("## ERSDKDelegate onLogin: setNotificationToken failed due to \(error.localizedDescription)")
                    }
                }
            }
        }
    }
    
    func onLogout(error: RiiiverSDK.ERError?) {
        // Notify to SDK that fcmToken should be teminated??
    }
    
    /// iiideaが実行開始した時に通知する
    /// - Parameters:
    ///   - applet: 起動したiiideaのインスタンス
    ///   - error: RiiiverSDKから返却されたerror情報
    func startApplet(applet: ERApplet, error: ERError?) {
        print("startApplet : appletId=\(applet.getId()), appletTitle=\(applet.metaData.getTitle()), sdkError=\(error?.errorMessage ?? "no error")" )
    }

    /// iiideaが実行終了した時に通知する
    /// - Parameters:
    ///   - applet: 終了したiiideaのインスタンス
    ///   - error: iiideaが正しく終了しなかった時のエラー
    ///   - reachBlock: iiideatが終了した時の最終到達Block
    ///   - output: iiideaが正しく終了した際の、最後に到達したPieceが出力したデータ
    func endApplet(applet: ERApplet, blockName: ERBlockName, output: Dictionary<String, Any>?, error: ERError?) {
        print("endApplet : appletId=\(applet.getId()), appletTitle=\(applet.metaData.getTitle())")
        if applet.isQCC5100Trigger, let error {
            // Check BLE Queue
            guard let queue = BLECentralClient.sharedInstance.entity.phoneToWatchQue else {
                print("## QCC5100ActionPieceCore: queue missing")
                return
            }

            let outputString = "iiidea failed due to \(error.errorMessage)"

            // Generate BLERequest
            var request: BLEOperation
            switch queue {
            case .button1result:
                request = BLEAPI.SendResultButton1(message: outputString)
            case .button2result:
                request = BLEAPI.SendResultButton2(message: outputString)
            case .display: fallthrough
            case .success: fallthrough
            case .failure: fallthrough
            default:
                print("## QCC5100ActionPieceCore: unhandlable queue found")
                return
            }

            // Invoke BLERequest
            BLECentralClient.sharedInstance.operate(request: request)
        }
    }
}
