//
//  DeviceSearchResultPresenterTest.swift
//  qcc5100.sampleTests
//

import XCTest
import CoreBluetooth
@testable import qcc5100_sample

class DeviceSearchResultPresenterTests: XCTestCase {
    
    var viewDelegate: ViewDelegateMock!
    var router: RouterMock!
    var searchDeviceInteractor: InteractorMock!
    var getAppInfoInteractor: GetAppInfoInteractorMock!
    var presenter: DeviceSearchResultPresenter!
    
    override func setUp() {
        super.setUp()
        
        viewDelegate = .init()
        router = .init()
        searchDeviceInteractor = .init()
        getAppInfoInteractor = .init()
        presenter = .init(
            viewDelegate: viewDelegate,
            router: router,
            searchDeviceInteractor: searchDeviceInteractor,
            getAppInfoInteractor: getAppInfoInteractor
        )
    }
    
    override func tearDown() {
    }
    
    func test_didSelect() {
        XCTContext.runActivity(named: "didSelect") { _ in
            XCTContext.runActivity(named: "before called") { _ in
                XCTContext.runActivity(named: "view delegate: `toggleOverlay` is not called") { _ in
                    XCTAssertEqual(viewDelegate.callCount_toggleOverlay, 0)
                }
                XCTContext.runActivity(named: "searchDeviceInteractor: `connect` is not called") { _ in
                    XCTAssertEqual(searchDeviceInteractor.callCount_connect, 0)
                }
            }
            
            XCTContext.runActivity(named: "after called: if the target is available") { _ in
                presenter.didSelect(peripheral: mockPeripheral)
                
                // "searchDeviceInteractor: `connect` is called"
                XCTAssertEqual(searchDeviceInteractor.callCount_connect, 1)
                
                // "view delegate: `toggelOverlay` is called twice"
                XCTAssertEqual(viewDelegate.callCount_toggleOverlay, 2)
                
                // "router: `segueToSetting` is called"
                XCTAssertEqual(router.isCalled_segueToSettings, true)
            }
            
            XCTContext.runActivity(named: "after called: if the target is unavailable") { _ in
                setUp()
                searchDeviceInteractor = .init(emulateUnavailable: true)
                presenter = .init(
                    viewDelegate: viewDelegate,
                    router: router,
                    searchDeviceInteractor: searchDeviceInteractor,
                    getAppInfoInteractor: getAppInfoInteractor
                )
                presenter.didSelect(peripheral: mockPeripheral)
                
                // "searchDeviceInteractor: `connect` is called"
                XCTAssertEqual(searchDeviceInteractor.callCount_connect, 1)
                
                // "view delegate: `toggelOverlay` is called twice"
                XCTAssertEqual(viewDelegate.callCount_toggleOverlay, 2)
                
                // "searchDeviceInteractor: `disconnectPeripheral` is called"
                XCTAssertEqual(searchDeviceInteractor.callCount_disconnectPeripheral, 1)
                
                // "view delegate: `showAlertUnavailable` is called")
                XCTAssertEqual(viewDelegate.callCount_showAlertUnavailable, 1)
            }
        }
    }
    
    func test_didTapInfo() {
        XCTContext.runActivity(named: "didTapInfo") { _ in
            presenter.didTapInfo()
            XCTAssertTrue(getAppInfoInteractor.isCalled_didTapInfo)
            XCTAssertEqual(viewDelegate.callCount_showAppInfo, 1)
            XCTAssertEqual(viewDelegate.appName, "app_name")
            XCTAssertEqual(viewDelegate.appBuild, "20240215.01")
            XCTAssertEqual(viewDelegate.appVersion, "1.0.0")
        }
    }
    
    
    // MARK: - Mock
    
    private let mockPeripheral = PeripheralEntity(
        name: "RIIIVER_DEVICE",
        uuid: UUID(uuidString: "acde070d-8c4c-4f0d-9d8a-162843c10333")!,
        advertisementData: ["hoge": "fuga"],
        rssi: 21,
        rawInstance: nil
    )
    
    class ViewDelegateMock: DeviceSearchResultViewInterface {
        var callCount_toggleOverlay = 0
        func toggleOverlay() {
            callCount_toggleOverlay += 1
        }

        var callCount_showAlertUnavailable = 0
        func showAlertUnavailable() {
            callCount_showAlertUnavailable += 1
        }
        
        var callCount_showAppInfo = 0
        var appName: String = ""
        var appBuild: String = ""
        var appVersion: String = ""
        func showAppInfo(_ info: qcc5100_sample.AppInfo) {
            callCount_showAppInfo += 1
            appName = info.name
            appBuild = info.build
            appVersion = info.version
        }
    }
    
    class RouterMock: DeviceSearchResultWireframe {
        var isCalled_segueToSettings = false
        func segueToSettings() {
            isCalled_segueToSettings = true
        }
    }
    
    class InteractorMock: SearchDeviceUsecase {
        init(emulateUnavailable: Bool = false) {
            if emulateUnavailable {
                self.isSlateRiiiver = false
            }
        }
        
        var isCalled_didTapScanButton = false
        func startScan() {
            isCalled_didTapScanButton.toggle()
        }
        
        var callCount_connect = 0
        var isSlateRiiiver: Bool = true
        func connect(to peripheral: PeripheralEntity, completion: @escaping ((Bool) -> Void)) {
            callCount_connect += 1
            completion(isSlateRiiiver)
        }
        
        var callCount_disconnectPeripheral = 0
        func disconnectPeripheral() {
            callCount_disconnectPeripheral += 1
        }
        
        var isCalled_deleteFoundPeripherals = false
        func deleteFoundPeripherals() {
            isCalled_deleteFoundPeripherals.toggle()
        }
    }
    
    class GetAppInfoInteractorMock: GetAppInfoUsecase {
        var isCalled_didTapInfo = false
        func getInfo() -> AppInfo {
            isCalled_didTapInfo.toggle()
            return AppInfo(name: "app_name", build: "20240215.01", version: "1.0.0")
        }
    }
}
