//
//  DeviceSettingsView.swift
//  qcc5100.sample
//

import SwiftUI
import RiiiverSDK

protocol DeviceSettingsViewInterface {
    func toggleOverlay()
    func addOptions(_ newOption: PreferredFunctionEntity)
}

let dummyFunction1: PreferredFunctionEntity = PreferredFunctionEntity(displayName: "Ring the phone(dummy)")
let dummyFunction2: PreferredFunctionEntity = PreferredFunctionEntity(displayName: "Control music(dummy)")

final class DeviceSettingsViewDelegate: Equatable, ObservableObject, DeviceSettingsViewInterface {
    static func == (lhs: DeviceSettingsViewDelegate, rhs: DeviceSettingsViewDelegate) -> Bool {
        return lhs.isShowOverlay == rhs.isShowOverlay && lhs.formOptions == rhs.formOptions && lhs.selected1 == rhs.selected1 && lhs.selected2 == rhs.selected2
    }
    
    @Published var isShowOverlay: Bool = false
    @Published var formOptions: [PreferredFunctionEntity] = [dummyFunction1, dummyFunction2]
    @Published var selected1: String = dummyFunction1.displayName
    @Published var selected2: String = dummyFunction2.displayName
    
    init(isShowOverlay: Bool = false,
         formOptions: [PreferredFunctionEntity] = [dummyFunction1, dummyFunction2],
         selected1: String = dummyFunction1.displayName,
         selected2: String = dummyFunction2.displayName
    ) {
        self.isShowOverlay = isShowOverlay
        self.formOptions = formOptions
        self.selected1 = selected1
        self.selected2 = selected2
        updateSelected()
    }

    func toggleOverlay() {
        self.isShowOverlay.toggle()
    }
    
    func addOptions(_ newOption: PreferredFunctionEntity) {
        self.formOptions.append(newOption)
    }
    
    private func updateSelected() {
        if let button1IiideaId = RiiiverAppletsEntity.sharedInstance.slateAppletIds[.button1result],
           let option1 = formOptions.first(where: { $0.iiidea?.getId() == button1IiideaId }) {
            self.selected1 = option1.displayName
        }
        else {
            selected1 = dummyFunction1.displayName
        }
        if let button2iiideaId = RiiiverAppletsEntity.sharedInstance.slateAppletIds[.button2result],
           let option2 = formOptions.first(where: { $0.iiidea?.getId() == button2iiideaId }) {
            self.selected2 = option2.displayName
        }
        else {
            selected2 = dummyFunction2.displayName
        }
    }
}

struct DeviceSettingsView: View {
    @ObservedObject var dataSource: BLEEntity
    @ObservedObject var riiiverUserEntity: RiiiverAppletsEntity
    @StateObject private var delegate: DeviceSettingsViewDelegate
    
    var presenter: DeviceSettingsPresentation!

    init(delegate: DeviceSettingsViewDelegate) {
        self.dataSource = BLECentralClient.sharedInstance.entity
        self.riiiverUserEntity = RiiiverSDKClient.sharedInstance.entity
        self._delegate = StateObject(wrappedValue: delegate)
    }

    var body: some View {
        NavigationView {
            VStack(spacing: 20) {
                if dataSource.currentConnected == nil {
                    Text("Disconnected")
                } else {
                    Text("Connected to \(dataSource.currentConnected!.name)")
                    Image("button_description")
                        .resizable()
                        .scaledToFit()
                        .frame(height: 200)
                    HStack {
                        Text("Function for Button1")
                            .padding(.leading, 8)
                        Picker("", selection: $delegate.selected1) {
                            ForEach(delegate.formOptions) { option in
                                Text(option.displayName).tag(option.displayName)
                            }
                        }
                        .onChange(of: delegate.selected1) { newValue in
                            Task {
                                if let iiideaId = delegate.formOptions.first(where: { $0.displayName == delegate.selected1 })?.iiidea?.getId() {
                                    await presenter.onSubmit(of: .button1result, iiideaId: iiideaId)
                                }
                            }
                        }
                        Spacer()
                    }
                    HStack {
                        Text("Function for Button2")
                            .padding(.leading, 8)
                        Picker("", selection: $delegate.selected2) {
                            ForEach(delegate.formOptions) { option in
                                Text(option.displayName).tag(option.displayName)
                            }
                        }
                        .onChange(of: delegate.selected2) { newValue in
                            Task {
                                if let iiideaId = delegate.formOptions.first(where: { $0.displayName == delegate.selected2 })?.iiidea?.getId() {
                                    await presenter.onSubmit(of: .button2result, iiideaId: iiideaId)
                                }
                            }
                        }
                        Spacer()
                    }
                    HStack {
                        Button("Add function with Riiiver") { self.presenter.didTapRiiiver() }
                            .padding(.leading, 8)
                        Spacer()
                    }
                }
            }
            .navigationBarTitle("Device Detail", displayMode: .inline)
            .toolbar {
                ToolbarItem(placement: .navigationBarLeading) {
                    Button("Back") { self.presenter.didTapBack() }
                }
            }
            .onAppear {
                Task { [self] in
                    await presenter.didAppear()
                }
            }
        }
        .overlay {
            if delegate.isShowOverlay {
                ZStack {
                    Color.gray.opacity(0.3)
                    ProgressView()
                }
                .ignoresSafeArea()
            }
        }
    }
}

struct PreferredFunctionEntity: Hashable, Equatable, Identifiable {
    var id: String {
        return displayName
    }
    var displayName: String
    var iiidea: ERApplet?
    
    static func == (lhs: PreferredFunctionEntity, rhs: PreferredFunctionEntity) -> Bool {
        return lhs.displayName == rhs.displayName && lhs.iiidea == rhs.iiidea
    }
}
